/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class IndexScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;                      // 表名称
    TabMeta tab_;                               // 表的元数据
    std::vector<Condition> conds_;              // 扫描条件
    RmFileHandle *fh_;                          // 表的数据文件句柄
    std::vector<ColMeta> cols_;                 // 需要读取的字段
    size_t len_;                                // 选取出来的一条记录的长度
    std::vector<Condition> fed_conds_;          // 扫描条件，和conds_字段相同

    std::vector<std::string> index_col_names_;  // index scan涉及到的索引包含的字段
    IndexMeta index_meta_;                      // index scan涉及到的索引元数据

    Rid rid_;
    std::unique_ptr<RecScan> scan_;

    SmManager *sm_manager_;

    int indexscan_;

   public:
    IndexScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, std::vector<std::string> index_col_names,
                    Context *context,IndexMeta idx_meta) {
                        // std::cout<<"qu"<<std::endl;
        sm_manager_ = sm_manager;
        context_ = context;
        tab_name_ = std::move(tab_name);
        tab_ = sm_manager_->db_.get_table(tab_name_);
        // optmize conds
        for(auto& cond:conds) {
            if(get_col(tab_.cols,cond.lhs_col)->type!=cond.rhs_val.type) {
                if(get_col(tab_.cols,cond.lhs_col)->type==ColType::TYPE_STRING||
                    cond.rhs_val.type==ColType::TYPE_STRING)
                    throw IncompatibleTypeError(coltype2str(get_col(tab_.cols,cond.lhs_col)->type), coltype2str(cond.rhs_val.type));
                if(get_col(tab_.cols,cond.lhs_col)->type==ColType::TYPE_FLOAT) {
                    cond.rhs_val.set_float(static_cast<double> (cond.rhs_val.int_val));
                    cond.rhs_val.init_raw(sizeof(double));
                }
                else if(get_col(tab_.cols,cond.lhs_col)->type==ColType::TYPE_INT) {
                    if(cond.op==OP_GT || cond.op == OP_LE) {
                    cond.rhs_val.set_int(static_cast<int> (cond.rhs_val.float_val));
                    }
                    else {
                    cond.rhs_val.set_int(static_cast<int> (cond.rhs_val.float_val)+1);
                    }
                    cond.rhs_val.init_raw(sizeof(int));
                }
                else
                    throw IncompatibleTypeError(coltype2str(get_col(tab_.cols,cond.lhs_col)->type), coltype2str(cond.rhs_val.type));
            }   
        }
        conds_ = std::move(conds);
        // index_no_ = index_no;
        index_col_names_ = index_col_names; 
        index_meta_ = idx_meta;
        // std::cout<<"idx"<<idx_meta.col_num<<std::endl;
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab_.cols;
        len_ = cols_.back().offset + cols_.back().len;
        std::map<CompOp, CompOp> swap_op = {
            {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
        };

        for (auto &cond : conds_) {
            if (cond.lhs_col.tab_name != tab_name_) {
                // lhs is on other table, now rhs must be on this table
                assert(!cond.is_rhs_val && cond.rhs_col.tab_name == tab_name_);
                // swap lhs and rhs
                std::swap(cond.lhs_col, cond.rhs_col);
                cond.op = swap_op.at(cond.op);
            }
        }
        indexscan_ =conds_.size();
        for(int i=0; i<conds_.size();i++) {
            if(std::find(index_col_names_.begin(),index_col_names_.end(),conds_[i].lhs_col.col_name)
                ==index_col_names_.end()){
                    indexscan_ = i;
                    break;
            }
        }
        // std::cout<<"indexscan"<<indexscan_<<std::endl;
        if(indexscan_!=conds_.size()) {
            fed_conds_.assign(conds_.begin()+indexscan_,conds_.end());
            for(auto&cond:fed_conds_) {
                // std::cout<<"fedconds:"<<cond.lhs_col.col_name<<std::endl;
            }
        }

    }

    void beginTuple() override {
        // std::cout<<"hahahaha"<<std::endl;
        IxIndexHandle *ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_,index_meta_)).get();
        // std::cout<<"name"<<std::endl;
        Iid leaf_lower = ih->leaf_begin();
        Iid leaf_upper = ih->leaf_end();
        // printf("start iid page %d, slot %d, end iid page %d, slot %d\n", 
        //     leaf_lower.page_no, leaf_lower.slot_no, leaf_upper.page_no, leaf_upper.slot_no);
        //map(cond.colname, cond)

        // std::cout<<"name"<<std::endl;
        // std::unordered_map<std::string, Condition> nameToConditionMap;
        // std::cout<<"name"<<std::endl;
        // for(int i = 0; i<index_col_names_.size(); i++) {
        //     std::cout<<"name"<<index_col_names_[i]<<std::endl;
        // }
        // for(auto& cond:conds_) {
        //     std::cout<<"shenren"<<std::endl;
        //     nameToConditionMap[cond.lhs_col.col_name] = cond;
        // }
        //使fedcond顺序与索引顺序一直
        // for(int i = 0; i<index_col_names_.size(); i++) {
        //     fed_conds_[i] = nameToConditionMap[index_col_names_[i]];
        // }
        

        //create key
        char *key = new char[index_meta_.col_tot_len];
        int offset = 0;
        int eq_conds_size = 0;

        for(int i = 0; i<index_col_names_.size()-1; i++) {
            // std::cout<<conds_.size()<<std::endl;
            auto cond = conds_[i];
            if(cond.op==OP_EQ) {
                memcpy(key+offset, cond.rhs_val.raw->data, cond.rhs_val.raw->size);
                offset += cond.rhs_val.raw->size;
                eq_conds_size++;
            }
            else {
                break;
            }
        }
        // printf("idxsacn %d, eqcondsz %d\n", indexscan_, eq_conds_size);
        if(indexscan_-eq_conds_size>1) {
            std::vector<Condition> last_conds (conds_.begin()+eq_conds_size,conds_.begin()+indexscan_);
            // std::cout<<"wonengwenwenma"<<last_conds.size()<<std::endl;
            for(auto& last_cond:last_conds) {
                int total_cols = eq_conds_size+1;
                // std::cout<<"wonengwenwenma"<<total_cols<<std::endl;
                switch (last_cond.op)
                {
                case OP_GE:
                    memcpy(key+offset, last_cond.rhs_val.raw->data, last_cond.rhs_val.raw->size);
                    //todo: bound函数重载
                    leaf_lower = ih->lower_bound(key, total_cols);
                    break;
                case OP_GT:
                    memcpy(key+offset, last_cond.rhs_val.raw->data, last_cond.rhs_val.raw->size);
                    //todo: bound函数重载
                    leaf_lower = ih->upper_bound(key, total_cols);
                    break;
                case OP_LE:
                    memcpy(key+offset, last_cond.rhs_val.raw->data, last_cond.rhs_val.raw->size);
                    //todo: bound函数重载
                    leaf_upper = ih->upper_bound(key, total_cols);
                    break;
                case OP_LT:
                    memcpy(key+offset, last_cond.rhs_val.raw->data, last_cond.rhs_val.raw->size);
                    //todo: bound函数重载
                    leaf_upper = ih->lower_bound(key, total_cols);
                    break;
                default:
                    throw UnknownOperatorError();
                    break;
                }
            }
        }
        else if(conds_.size()!=0){
            // std::cout<<"hahahahsa"<<std::endl;

              //根据最后一个fed_cond确定范围
            auto last_cond = conds_[indexscan_-1];
            int cols = indexscan_-1;
            // std::cout<<"op"<<std::endl;
            // std::cout<<cols<<std::endl;
            switch (last_cond.op) {
                case OP_EQ:
                    // std::cout<<"op"<<std::endl;
                    // std::cout<<last_cond.rhs_val.int_val<<std::endl;
                    memcpy(key+offset, last_cond.rhs_val.raw->data, last_cond.rhs_val.raw->size);
                    cols ++; //todo: bound函数重载
                    leaf_lower = ih->lower_bound(key,cols);   
                    //leaf_lower = ih->lower_bound(key,offset);   
                    leaf_upper = ih->upper_bound(key,cols);
                    
                    break;
                case OP_GE:
                    leaf_upper = ih->upper_bound(key,cols);
                    memcpy(key+offset, last_cond.rhs_val.raw->data, last_cond.rhs_val.raw->size);
                    cols++;//todo: bound函数重载
                    leaf_lower = ih->lower_bound(key,cols);
                    break;
                case OP_GT:
                    leaf_upper = ih->upper_bound(key,cols);
                    memcpy(key+offset, last_cond.rhs_val.raw->data, last_cond.rhs_val.raw->size);
                    cols++;//todo: bound函数重载
                    leaf_lower = ih->upper_bound(key,cols);
                    break;
                case OP_LE:
                    leaf_lower = ih->lower_bound(key,cols);
                    memcpy(key+offset, last_cond.rhs_val.raw->data, last_cond.rhs_val.raw->size);
                    cols++;//todo: bound函数重载
                    leaf_upper = ih->upper_bound(key,cols);
                    break;
                case OP_LT:
                    leaf_lower = ih->lower_bound(key,cols);
                    memcpy(key+offset, last_cond.rhs_val.raw->data, last_cond.rhs_val.raw->size);
                    cols++;//todo: bound函数重载
                    leaf_upper = ih->lower_bound(key,cols);
                    break;
                default:
                    throw UnknownOperatorError();
                    break;
            }
        }
        
        scan_ = std::make_unique<IxScan>(ih,leaf_lower,leaf_upper, sm_manager_->get_bpm());  // 使用具体实现
        // printf("start iid page %d, slot %d, end iid page %d, slot %d\n", 
        //     leaf_lower.page_no, leaf_lower.slot_no, leaf_upper.page_no, leaf_upper.slot_no);
        if(fed_conds_.size()>0) {
            // std::cout<<"begin"<<std::endl;
            while (!scan_->is_end()) {
            // std::cout<<"begin"<<std::endl;
            rid_=scan_->rid();
            if(!check_conditions(fh_->get_record(rid_, context_))) {
                scan_->next();
            }
            else {
                return;
            }
            }
        }
        else {
            if(!scan_->is_end()) {
                // std::cout<<rid_.page_no;
                rid_ = scan_->rid();
            }
        }

    }

    void nextTuple() override {
        // printf("nexttuple\n");
        if(fed_conds_.size()>0) {
            scan_->next();
            while (!scan_->is_end()) {
            // std::cout<<"next"<<std::endl;
            rid_ = scan_->rid();
            std::unique_ptr<RmRecord> record = fh_->get_record(rid_, context_);
            if (check_conditions(record)) {
                return;
            }
            scan_->next();
            }
        }
        else {
            scan_->next();
            if(!scan_->is_end()) {
                rid_ = scan_->rid();
            }  
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        if (scan_->is_end()) {
            return nullptr;
        }
        std::unique_ptr<RmRecord> record = fh_->get_record(rid_, context_);
        return record;
    }

    bool check_conditions(const std::unique_ptr<RmRecord> &record) {
        bool check = true;
        for (const auto &cond : fed_conds_) {
            auto lhs_col = get_col(cols_, cond.lhs_col);
            char *lhs_data = record->data + lhs_col->offset;
            char *rhs_data = nullptr;
            ColType rhs_type;
            if(cond.is_rhs_val) {
                rhs_data = cond.rhs_val.raw->data;
                rhs_type = cond.rhs_val.type;
            }else {
                auto rhs_col = get_col(cols_, cond.rhs_col);
                rhs_type = rhs_col->type;
                rhs_data = record->data + rhs_col->offset;
            }
            int cmp = 0;
            if (rhs_type == lhs_col->type) {
                cmp = ix_compare(lhs_data, rhs_data, rhs_type, lhs_col->len);
            }
            else {
                if (rhs_type == ColType::TYPE_STRING || lhs_col->type == ColType::TYPE_STRING) {
                    throw IncompatibleTypeError(coltype2str(lhs_col->type), coltype2str(rhs_type));
                }
                double lf = tofloat(lhs_data, lhs_col->type);
                double rf = tofloat(rhs_data, rhs_type);
                double diff = lf - rf;
                // std::cout<<lf<<std::endl<<rf<<std::endl<<diff<<std::endl;
                const double epsilon = 1e-7f;
                if (std::abs(diff) < epsilon) {
                    cmp = 0;
                } else {
                    cmp = (diff < 0) ? -1 : 1;
                }
            }
            switch (cond.op) {
            case OP_EQ:
                check = check && (cmp == 0);
                break;
            case OP_NE:
                check = check && (cmp != 0);
                break;
            case OP_LT:
                check = check && (cmp < 0);
                break;
            case OP_GT:
                check = check && (cmp > 0);
                break;
            case OP_LE:
                check = check && (cmp <= 0);
                break;
            case OP_GE:
                check = check && (cmp >= 0);
                break;
            default:
                throw UnknownOperatorError();
            }
        }
        return check;
    }
    double tofloat(char* data, ColType type) {
        switch (type) {
        case TYPE_INT: {
            // 将整型数据转换为 float
            int int_val;
            std::memcpy(&int_val, data, sizeof(int));
            return static_cast<float>(int_val);
        }
        case TYPE_FLOAT: {
            // 将浮点型数据直接转换为 float
            double float_val;
            std::memcpy(&float_val, data, sizeof(double));
            return float_val;
        }
        default:
            throw std::invalid_argument("Unsupported column type");
        }
    }

    Rid &rid() override { return rid_; }

    const std::vector<ColMeta> &cols() const override { return cols_; }

    size_t tupleLen() const override { return len_; }

    bool is_end() const override { return scan_->is_end(); }

    std::string getType() override {
        return "IndexScanExecutor";
    }

};